/*
 * include/linux/low-latency.h
 *
 * Andrew Morton <akpm@zip.com.au>
 */

#ifndef LOW_LATENCY_H_INCLUDED
#define LOW_LATENCY_H_INCLUDED

#define LOWLATENCY_NEEDED	1


#include <linux/cache.h>		/* For ____cacheline_aligned */

#define enable_lowlatency 1

/*
 * Set this non-zero to generate low-latency instrumentation
 */
#define LOWLATENCY_DEBUG		0

#define conditional_schedule_needed() (current->need_resched)

void set_running_and_schedule(void);
#define unconditional_schedule() set_running_and_schedule()

#define conditional_schedule()						\
	do {								\
		if (conditional_schedule_needed())			\
			unconditional_schedule();			\
	} while (0)

#define DEFINE_RESCHED_COUNT	int resched_count = 0
#define TEST_RESCHED_COUNT(n)	(enable_lowlatency && (++resched_count > (n)))
#define RESET_RESCHED_COUNT()	resched_count = 0
extern int ll_copy_to_user(void *to_user, const void *from, unsigned long len);
extern int ll_copy_from_user(void *to, const void *from_user, unsigned long len);


#endif /* LOW_LATENCY_H_INCLUDED */

